/*
 * Copyright (c) 2006, Adam Dunkels
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */
 ///////////////////////////////////////
 // modified for nonbasic syntax
 //
 // Bruce Land, Cornell Univ, May 2010
 // port to PIC32 August 2015
 ///////////////////////////////////////

#include <plib.h>
#include <inttypes.h>
#include <stdio.h>
#include "ubasic.h"
#define _SUPPRESS_PLIB_WARNING

#pragma config FNOSC = FRCPLL, POSCMOD = HS, FPLLIDIV = DIV_2, FPLLMUL = MUL_20, FPBDIV = DIV_1, FPLLODIV = DIV_2
#pragma config FWDTEN = OFF
// frequency we're running at
#define	SYS_FREQ 40000000

// UART parameters
#define BAUDRATE 9600 // must match PC end
#define PB_DIVISOR (1 << OSCCONbits.PBDIV) // read the peripheral bus divider, FPBDIV
#define PB_FREQ SYS_FREQ/PB_DIVISOR // periperhal bus frequency


static const char prog[] =
"print 'starting demo'; "
"for i= 0 to 10 { "
   "print 'i=', i ;"
   "for j=1 to 2 {print 'i-j=', i-j;};"
"};"
"print 'done' ; "
"end; "
;

// 10,000 nested loops in 6 seconds 
// (with peek, poke)
// just the poke is 4 seconds
// just loop 2 seconds so 0.2 mSec/loop
// so loop time is 
// adding k=k+1; 8.5 sec, 
// so 2.5/10000=0.25 millseconds for add
/*
static const char program[] = 
"print 'start' ;\
poke 7+32, 255 ; \
for i = 1 to 100 { \
  for j = 1 to 100 { \
   poke 8+32, i ; \
  }; \
}; \
print 'done' ;\
end; \
" ;


static const char program[] = 
"print 'start' ; \
for i = 1 to 10000 { j=i/5; } ; \
print 'end' ; \
end ; \
" ;
*/

//************************************************
static char rxchar='0'; 		//received character
int num ;
#define str_buffer_size 20
char str_buffer[str_buffer_size];
/*********************************/
// useful ASCII/VT100 macros for PuTTY
#define clrscr() printf( "\x1b[2J")
#define home()   printf( "\x1b[H")
#define pcr()    printf( '\r')
#define crlf     putchar(0x0a); putchar(0x0d);
#define backspace 0x08
//#define max_chars 30 // for input buffer
//////////////////////////////////////////////
// build a string from the UART2 /////////////
//////////////////////////////////////////////
int GetDataBuffer(char *buffer, int max_size)
{
    int num_char;
    num_char = 0;
    memset(buffer, 0, max_size);
    
    while(num_char < max_size)
    {
        char character;

        // get the character
        while(!UARTReceivedDataIsAvailable(UART2)){};
        character = UARTGetDataByte(UART2);
        UARTSendDataByte(UART2, character);

        // end line
        if(character == '\r'){
            *buffer = 0; // zero terminate the string
            crlf; // send a new line
            break;
        }
        // backspace
        if (character == backspace){
            putchar(' '); // write a blank over the previous character
            putchar(backspace); // go back one position
            buffer--; 
            num_char--; 
        }
        else { // save it all
            *buffer = character;
            buffer++;
            num_char++;
        } //if (character == backspace)

    } //while(num_char < max_size)
    return num_char;
}
/*---------------------------------------------------------------------------*/
int
main(void)
{
  char in_count, current_char;
  
  // specify PPS group, signal, logical pin name
    PPSInput (2, U2RX, RPB11); //Assign U2RX to pin RPB11 -- Physical pin 22 on 28 PDIP
    PPSOutput(4, RPB10, U2TX); //Assign U2TX to pin RPB10 -- Physical pin 21 on 28 PDIP
    // init the uart2
    UARTConfigure(UART2, UART_ENABLE_PINS_TX_RX_ONLY);
    UARTSetLineControl(UART2, UART_DATA_SIZE_8_BITS | UART_PARITY_NONE | UART_STOP_BITS_1);
    UARTSetDataRate(UART2, PB_FREQ, BAUDRATE);
    UARTEnable(UART2, UART_ENABLE_FLAGS(UART_PERIPHERAL | UART_RX | UART_TX));
    
  printf("%s\n", prog);
  
  // run the program 
  ubasic_init(prog);
  

  do {
    ubasic_run();
  } while(!ubasic_finished());
  
  
  return 0;
}
/*---------------------------------------------------------------------------*/
